// namespace
app = {};

/**
 * スタイルシートのクラス名に使用する分別区分の英字文字列取得メソッド。
 * 
 * @param {string} sortText 分別区分の日本語文字列。
 * @return {string} 分別区分の英字文字列。
 */
app.sortKey = sortText => {
    return {
// #################################################
/**
 * 分別区分を変更する場合は、この部分を修正してください。
 * 左側の値（Key）には、分別区分の日本語テキストを記載し、
 * 右側の値（Value）には、スタイルシートのクラス名に用いる
 * アルファベットのテキストを記載してください。
 * この関数により生成されるスタイルクラスは、次の4種類です。
 * ".catalog-list-item-sort--[text]"
 * ".dialog--[text]"
 * ".dialog-catalog-sort--[text]"
 * ".calendar-table-date-sort--[text]"
 */
        '可燃ごみ' : 'kanen',
        '不燃ごみ' : 'hunen',
        '資源ごみ' : 'shigen',
        '収集不可' : 'huka'
// #################################################
    }[sortText];
};

/**
 * CSV取得メソッド
 * @param {*} text 
 * @returns csvテキスト(Promiseオブジェクト)
 */
app.getCSV = (url) => {
    return fetch(url).then(response => {
        return response.text();
    });
};

/**
 * CSVテキストを1行ずつ読み取るメソッド。
 * 読み取ったデータはコールバック関数で処理する。
 * 
 * @param {string} csv 
 * @param {number} start 
 * @param {callback} handler 
 */
app.readCSV = (csv, start, handler) => {
    // csvの最終行の改行を削除
    csv = csv.replace(/\r\n$/, '');
    // 行(\r\n)に切り分け
    let lines = csv.split(/\r\n(?=(?:[^"]*"[^"]*")*[^"]*$)/);
    for (i = start; i < lines.length; i++) {
        // 値に切り分け(カンマ区切り)
        const values = lines[i].split(/,(?=(?:[^"]*"[^"]*")*[^"]*$)/g).map(value => {
            // ダブルクオート削除
            return value.replace(/^"((?:[^"]*"[^"]*"[^"]*)|[^"]*)"$/, (m, p) => { return p || m });
        });
        handler(i, values);
    };
};

/**
 * ひらがな変換メソッド
 * 
 * @param {*} text 
 * @returns 
 */
app.convertToHiragana = text => {
    if (typeof text != 'string') return '';
    return text.replace(/[\u30a1-\u30f6]/g, match => String.fromCharCode(match.charCodeAt(0) - 0x60));
};

/**
 * カタカナ変換メソッド
 * 
 * @param {*} text 
 * @returns 
 */
 app.convertToKatakana = text => {
    if (typeof text != 'string') return '';
    return text.replace(/[\u3041-\u3096]/g, match => String.fromCharCode(match.charCodeAt(0) + 0x60));
};

/**
 * 半角カナの全角変換メソッド
 * 
 * @param {*} text 
 * @returns 
 */
 app.convertToZenkaku = text => {
    if (typeof text != 'string') return '';
    const kanaMap = {
        'ｶﾞ': 'ガ', 'ｷﾞ': 'ギ', 'ｸﾞ': 'グ', 'ｹﾞ': 'ゲ', 'ｺﾞ': 'ゴ',
        'ｻﾞ': 'ザ', 'ｼﾞ': 'ジ', 'ｽﾞ': 'ズ', 'ｾﾞ': 'ゼ', 'ｿﾞ': 'ゾ',
        'ﾀﾞ': 'ダ', 'ﾁﾞ': 'ヂ', 'ﾂﾞ': 'ヅ', 'ﾃﾞ': 'デ', 'ﾄﾞ': 'ド',
        'ﾊﾞ': 'バ', 'ﾋﾞ': 'ビ', 'ﾌﾞ': 'ブ', 'ﾍﾞ': 'ベ', 'ﾎﾞ': 'ボ',
        'ﾊﾟ': 'パ', 'ﾋﾟ': 'ピ', 'ﾌﾟ': 'プ', 'ﾍﾟ': 'ペ', 'ﾎﾟ': 'ポ',
        'ｳﾞ': 'ヴ', 'ﾜﾞ': 'ヷ', 'ｦﾞ': 'ヺ',
        'ｱ': 'ア', 'ｲ': 'イ', 'ｳ': 'ウ', 'ｴ': 'エ', 'ｵ': 'オ',
        'ｶ': 'カ', 'ｷ': 'キ', 'ｸ': 'ク', 'ｹ': 'ケ', 'ｺ': 'コ',
        'ｻ': 'サ', 'ｼ': 'シ', 'ｽ': 'ス', 'ｾ': 'セ', 'ｿ': 'ソ',
        'ﾀ': 'タ', 'ﾁ': 'チ', 'ﾂ': 'ツ', 'ﾃ': 'テ', 'ﾄ': 'ト',
        'ﾅ': 'ナ', 'ﾆ': 'ニ', 'ﾇ': 'ヌ', 'ﾈ': 'ネ', 'ﾉ': 'ノ',
        'ﾊ': 'ハ', 'ﾋ': 'ヒ', 'ﾌ': 'フ', 'ﾍ': 'ヘ', 'ﾎ': 'ホ',
        'ﾏ': 'マ', 'ﾐ': 'ミ', 'ﾑ': 'ム', 'ﾒ': 'メ', 'ﾓ': 'モ',
        'ﾔ': 'ヤ', 'ﾕ': 'ユ', 'ﾖ': 'ヨ',
        'ﾗ': 'ラ', 'ﾘ': 'リ', 'ﾙ': 'ル', 'ﾚ': 'レ', 'ﾛ': 'ロ',
        'ﾜ': 'ワ', 'ｦ': 'ヲ', 'ﾝ': 'ン',
        'ｧ': 'ァ', 'ｨ': 'ィ', 'ｩ': 'ゥ', 'ｪ': 'ェ', 'ｫ': 'ォ',
        'ｯ': 'ッ', 'ｬ': 'ャ', 'ｭ': 'ュ', 'ｮ': 'ョ',
        '｡': '。', '､': '、', 'ｰ': 'ー', '｢': '「', '｣': '」', '･': '・'
    };
    const reg = new RegExp(`(${Object.keys(kanaMap).join('|')})`, 'g');
    return text.replace(reg, match => kanaMap[match]);
},

/**
* スクロールアニメーション
 * 
 * @param {element} element スクロールする要素。 
 * @param {number} amount スクロール量。
 * @param {number} freq スクロールのスピード、初期値20。
 */
app.animateScroll = (element, amount, freq = 20) => {
    const start = element.scrollTop;
    let i = 0;
    const easeOut = p => p * (2 - p);
    const scroll = timestamp => {
        element.scrollTo(0, start + (amount * easeOut(i / freq)));
        if (++i <= freq) requestAnimationFrame(scroll);
    };
    requestAnimationFrame(scroll);
};


/*
 * ダイアログの表示用メソッド
 */
app.showDialog = (html, options = {}) => {
    const dialog = document.getElementById('dialog');
    app.dialog.options = options;
    if (options.hasOwnProperty('class') && options.class) dialog.classList.add(options.class);
    dialog.querySelector('.dialog-content').innerHTML = html;
    dialog.classList.add('dialog--show');
    document.getElementById('dialog-dismiss').classList.add('dialog-dismiss--show');
};

/*
 * ダイアログを閉じるメソッド
 */
app.closeDialog = () => {
    const options = app.dialog.options;
    document.getElementById('dialog').classList.remove('dialog--show');
    document.getElementById('dialog-dismiss').classList.remove('dialog-dismiss--show');
    // 初期化
    setTimeout(() => {
        if (options.hasOwnProperty('class') && options.class) {
            dialog.classList.remove(options.class);
            dialog.options = {};
        };    
    }, 100);
};

/*
 * ダイアログの初期化メソッド
 */
app.initDialog = () => {
    app.dialog = {};
    app.dialog.options = {};
        // ダイアログを閉じるイベントをセット
    document.querySelector('.dialog-close').addEventListener('click', app.closeDialog);
    document.getElementById('dialog-dismiss').addEventListener('click', app.closeDialog);
};

/**
 * 外部ファイルの読み込みメソッド。
 * 読み込み完了後にPromiseオブジェクトを返す。
 * 
 * @param {string} path ファイルパス。
 * @param {string} type ファイルのタイプ （"js" or "css"）。
 * @return Promise オブジェクト。
 */
app.import = (path) => {
    let element;
    let loaded = false;
    const type = path.match(/(?!\.)[^\.]*$/).toString().toLowerCase();
    switch (type) {
        case 'js':
            element = document.createElement('script');
            element.src = path;
            element.addEventListener('load', e => {
                loaded = true;
            });
            break;

        case 'css':
            element = document.createElement('link');
            element.rel = 'stylesheet';
            element.href = path;
            element.addEventListener('load', e => {
                loaded = true;
            });
            break;
    };

    // DOM に追加して読み込み
    document.body.appendChild(element);

    // 読み込み完了後にPromiseオブジェクトを返す
    return new Promise(resolve => {
        const int = setInterval(() => {
            if (loaded) {
                clearInterval(int);
                return resolve(true);
            };
        }, 10); // 最低10ms待つ
    });

};

/**
 * コンテンツの切り替え処理。
 * 最初に表示する前に、jsとcssを読み込み初期化する。
 * 
 * @param {string} contentId コンテンツを識別するID（"catalog"|"calendar"|"place"|"help"|"gsi"）。
 */
app.switchContet = async (contentId) => {

    // 表示中のコンテンツを取得
    const currentContent = document.querySelector('.content--show');

    // コンテンツが表示中の場合は処理終了
    if (currentContent && currentContent.id == contentId) return;

    // 表示用クラスの削除と付与
    if (currentContent) currentContent.classList.remove('content--show');
    document.getElementById(contentId).classList.add('content--show');
    
    // コンテンツの初期化（初回のみ）
    if (!(contentId in app)) {
        switch (contentId) {
            case 'catalog':
                // 初期化フラグ
                app[contentId] = true;
                // 初期化実行
                app.initCatalog();
                break;

            case 'calendar':
                app[contentId] = true;
                app.initCalendar();
                break;

            case 'place':
                app[contentId] = true;
                app.initPlace();
                break;
            
            case 'help':
                app[contentId] = true;
                app.initHelp();
                break;

            case 'gsi':
                app[contentId] = true;
                app.initGsi();
                break;
        };
    };

};

/**
 * メニューの初期化メソッド
 */
app.initMenu = () => {

    // メニューボタンにメニュー表示/非表示切り替え処理をセット
    document.querySelector('.header-menu-button').addEventListener('click', e => {
        const menu = document.querySelector('.header-menu');
        const dismiss = document.querySelector('.menu-dismiss');
        if (!menu.classList.contains('header-menu--show')) {
            e.currentTarget.classList.add('header-menu-button--active');
            menu.classList.add('header-menu--show');
            dismiss.classList.add('menu-dismiss--show');
        } else {
            e.currentTarget.classList.remove('header-menu-button--active');
            menu.classList.remove('header-menu--show');
            dismiss.classList.remove('menu-dismiss--show');
        }
    });
    // dismiss要素のクリックイベント
    document.querySelector('.menu-dismiss').addEventListener('click', e => {
        document.querySelector('.header-menu-button').classList.remove('header-menu-button--active');
        document.querySelector('.header-menu').classList.remove('header-menu--show');
        e.currentTarget.classList.remove('menu-dismiss--show');
    });  

    // メニュー項目のクリックイベントをセット
    document.querySelectorAll('.header-menu-item[data-content-id]').forEach(item => {
        
        item.addEventListener('click', e => {

            // メニューの選択クラスの削除と付与
            document.querySelectorAll('.header-menu-item--selected').forEach(item => {
                item.classList.remove('header-menu-item--selected');
            });
            e.currentTarget.classList.add('header-menu-item--selected');

            // メニューダイアログを閉じる（メニューバー表示時は動いても影響ない）
            document.querySelector('.menu-dismiss').dispatchEvent(new Event('click'));

            // コンテンツの切り替え
            app.switchContet(e.currentTarget.dataset.contentId);

        });
    });
};

/**
 * ヘルプページの初期化メソッド
 */
app.initHelp = () => {

    // コンテンツリンク（URL変更あり）のクリックイベントをセット
    document.querySelectorAll('.content-link').forEach(elm => {
        const hash = new URL(elm.href).hash;
        const url = location.origin + location.pathname + hash;
        elm.href = url;

        // テキストも変更 --- 後で直接テキストに変える
        elm.innerText = url; 

        // ページ内リンクの強制リロード
        elm.addEventListener('click', e => {
            window.location.href = url;
            window.location.reload();
        });
    });

    // コンテンツリンク（URL変更なし）のクリックイベント
    document.querySelectorAll('.content-switch').forEach(elm => {
        elm.addEventListener('click', e => {
            const hash = new URL(e.currentTarget.href).hash;
            app.switchContet(hash.replace('#', ''));
        });
    });

};

/*
* 分別一覧ページの初期化メソッド
* 
*/
app.initCatalog = async () => {

    /**
     * ##############
     *  関数定義
     * ##############
     */

    app.catalog = {};
    app.catalog.data = [];

    /**
     * catalog.csv のデータを読み取ってオブジェクトに保存する。
     * 
     * @param {string} csv 
     * @return {object[]} ごみ品目リストのオブジェクト配列
     */
    const parseCatalogCsv = async (path) => {

        // CSVデータを取得
        const csv = await app.getCSV(path);

        // CSVの品目データをオブジェクトに保存
        let items = [];
        app.readCSV(csv, 1, (row, values) => {
            items.push({
                name        :values[0],
                kana        :values[1],
                keyword     :values[2],
                sort        :values[3],
                sortKey     :app.sortKey(values[3]),
                description :values[4],
                recycle     :values[5],
                linkTitle   :values[6],
                linkUrl     :values[7]
            });
        });
        return items;
    };

    /**
     * 
     * @param {object[]} data 品目データオブジェクト。 
     * @return {string} 品目リストのHTMLテキスト。
     */
    const createListItemHtml = data => {
        let htmls = [];
        for (let i = 0; i < data.length; i++) {
            const item = data[i];
            const sortClass = `catalog-list-item-sort--${app.sortKey(item.sort)}`;
            let keyword = `${item.name} ${item.kana} ${item.keyword} ${item.sort}`;
            keyword = keyword.replace(/\s?[,、]\s?/g, ' ').replace(/\snull/i, '');
            htmls.push(`
                <li class="catalog-list-item" data-index="${i}" data-kana="${item.kana}" data-keyword="${keyword}">
                    <h2 class="catalog-list-item-name">${item.name}</h2>
                    <h3 class="catalog-list-item-sort ${sortClass}">${item.sort}</h3>
                    <p class="catalog-list-item-description">${item.description}</p>
                </li>
            `);
        };
        return htmls.join('');
    };

    /**
     * 
     * @param {object} item 
     * @return {string} ダイアログに表示するHTML。
     */
    const createDialogHtml = item => {
        const sortKey = app.sortKey(item.sort);
        let html = `
            <h2 class="dialog-catalog-name">${item.name}</h2>
            <h3 class="dialog-catalog-sort dialog-catalog-sort--${sortKey}">${item.sort}</h3>
            <p class="dialog-catalog-description">${item.description.replace(/\r\n|\n/g, '<br>')}</p>
        `;
        if (item.recycle) {
            html += `<h3 class="dialog-catalog-recycle-title">リサイクル</h3>`;
            html += `<p class="dialog-catalog-recycle">${item.recycle.replace(/\r\n|\n/g, '<br>')}</p>`;
        }
        if (item.linkTitle && item.linkUrl) {
            html += `<div><a class="dialog-catalog-link" href="${item.linkUrl}" target="blank">${item.linkTitle}</a></div>`;
        }
        return html;
    };

    /**
     * 検索パターンを成形するメソッド。
     * 全角に変換し、ひらがなとカタカナのパターンを追加する。
     * 品目データは基本的に全角で入力されている前提なので、半角変換は行わない。
     * （例："ﾃｽﾄ" >> "ﾃｽﾄ|てすと|テスト"）
     * 
     * @param {string} pattern 
     * @return 正規表現のパターン文字列
     */    
    const adjustSearchPattern = pattern => {
        pattern = app.convertToZenkaku(pattern);
        pattern = pattern.trim();
        pattern = pattern.replace(/[\s]+/g, '.+');
        if (pattern.length > 0) {
            pattern = `${app.convertToHiragana(pattern)}|${app.convertToKatakana(pattern)}`;
        }
        return pattern;
    };

    /**
     * かなスクロールの初期化メソッド
     */
    const initScroller = () => {
        const scrollerItems = document.querySelectorAll('.catalog-scroller-item');

        // クリック時のイベントをセット
        scrollerItems.forEach(item => {
            item.addEventListener('click', e => {

                // disabled クラスが付与されている場合は処理を行わない
                if (e.currentTarget.parentElement.classList.contains('catalog-scroller--disabled')) return;

                const letters = e.currentTarget.dataset.letters;

                // 表示する要素を取得
                const target = [...document.querySelectorAll('.catalog-list-item')].find(item => {
                    return letters.includes(item.dataset.kana.substring(0, 1));
                });

                // スクロール
                const targetPos = target.offsetHeight * target.dataset.index - 4; // -4px は余白
                const scrollAmount = targetPos - target.parentElement.scrollTop;
                app.animateScroll(target.parentElement, scrollAmount);

            });
        });

        // リストのスクロールイベントをセット（現在位置をかなリストに反映）
        document.getElementById('catalog-list').addEventListener('scroll', e => {

            // 表示している最上段の要素の kana の1文字目を取得
            const list = e.currentTarget;
            const index = Math.ceil(list.scrollTop / list.children[0].offsetHeight);
            let char = list.children[index].dataset.kana.substring(0, 1);
            if (list.scrollTop + list.clientHeight + 20 > list.scrollHeight) char = "わ";
            
            // かなスクロールの状態を更新
            const currentClass = 'catalog-scroller-item--current';
            const currentItem = document.querySelector('.' + currentClass);
            if (!currentItem) return; // currentItem が取得できなかった場合は終了
            if (currentItem.dataset.letters.includes(char)) return; // 変更が必要ない場合は終了
            currentItem.classList.remove(currentClass);
            document.querySelector(`.catalog-scroller-item[data-letters*="${char}"]`).classList.add(currentClass);

        });
    };

    /**
     * ##############
     *  処理実行
     * ##############
     */


    // 品目データオブジェクトを生成
    const items = app.catalog.data = await parseCatalogCsv('data/catalog.csv');

    // 品目リストの中身要素を生成
    document.getElementById('catalog-list').innerHTML = createListItemHtml(items);

    // 品目リストの項目にイベントをセット
    document.querySelectorAll('.catalog-list-item').forEach(elm => {
        elm.addEventListener('click', e => {
            const item = items[e.currentTarget.dataset.index];
            const html = createDialogHtml(item);
            app.showDialog(html, {class: `dialog--${(item.sortKey)}`});
        });
    });

    // かなスクロールの初期化
    initScroller();

    // 検索ボックスのイベント（入力時）をセット
    document.getElementById('catalog-filter-input').addEventListener('input', e => {

        const input = e.currentTarget;

        // clear ボタンの表示切替
        // scroller の無効化
        const clearButton = document.querySelector('.catalog-filter-clear');
        const scroller = document.querySelector('.catalog-scroller');
        if (input.value.length) {
            clearButton.classList.remove('catalog-filter-clear--hidden');
            scroller.classList.add('catalog-scroller--disabled');
        } else {
            clearButton.classList.add('catalog-filter-clear--hidden');
            scroller.classList.remove('catalog-scroller--disabled');
        };

        // 項目要素の"data-keywords"属性の値と入力値を比較して表示を判定
        const pattern = adjustSearchPattern(input.value);
        const reg = new RegExp(pattern, 'i');
        document.querySelectorAll('.catalog-list-item').forEach(elm => {
            const keyword = elm.dataset.keyword;
            if (reg.test(keyword)) {
                elm.classList.remove('catalog-list-item--hidden');
            } else {
                elm.classList.add('catalog-list-item--hidden');
            }
        });
    });

    // 検索テキストクリアボタンのイベント
    document.querySelector('.catalog-filter-clear').addEventListener('click', e => {
        const input = document.querySelector('.catalog-filter-input');
        input.value = '';
        input.dispatchEvent(new Event('input'));
    });

};

/**
 * 収集日ページの初期化メソッド
 */
app.initCalendar = async () => {

    /**
     * ##############
     *  関数定義
     * ##############
     */

    /**
     * @param {object}   area           地区のデータを格納するオブジェクト
     * @param {number}   area.id        地区ID
     * @param {string}   area.name      地区名
     * @param {object[]} area.schedules schedule オブジェクトの配列。
     * 
     * @param {object}   schedule             収集日の情報を格納するオブジェクト
     * @param {string}   schedule.sort        分別区分
     * @param {string}   schedule.file        収集日CSVファイルの名前
     * @param {string[]} schedule.dateStrings 収集日の日付文字列（Date.toLocalDateString()）配列
     */
    app.areas;

    /**
     * CSVテキストをオブジェクト配列に変換する。
     * 
     * @param {string} path CSVファイルのパス。
     * @return {object[]} areaオブジェクトの配列。
     */
    const parseCalendarCsv = async (path) => {
        const csv = await app.getCSV(path);
        let headers;
        let areas = [];
        app.readCSV(csv, 0, (row, values) => {
            /**
             * values 配列の中身
             * [0] : 地区ID
             * [1] : 地区名
             * [2] : 可燃ごみ収集日のパターン (kanen1)
             * [3] : 不燃ごみ収集日のパターン (hunen1)
             * [4] : 資源ごみ収集日のパターン (shigen1)
             * ※ 収集日のパターンは（[2]以降）は順番変更も増減も可能
             */
    
            // 1行目（ヘッダー列を変数に保存）
            if (row == 0) {
                headers = values;
                return;
            };
    
            // 2行目以降
            const area = {
                id        : Number(values[0]),  /* 地区ID */
                name      : values[1],          /* 地区名 */
                schedules : []
            };
            for (let i = 2; i < values.length; i++) {
                area.schedules.push({
                    sort: headers[i],  // "可燃ごみ"
                    file: values[i].match(/\.csv$/i) ? values[i] : values[i] + '.csv',  // "kanen1.csv"
                    dateStrings: []    // 収集日文字列の配列（後で取得）
                });
            };
            areas.push(area);
        });
        return areas;
    };

    /**
     * 引数で与えた area オブジェクトの schedule.dateStrings プロパティを更新する。
     * データの取得には schedule.file の値を参照する。
     * 
     * @param {object} area この引数で与えたareaオブジェクトのプロパティを更新しする。
     */
    const loadSchedules = async (area) => {
    
        // 収集日データが既にある場合は処理中断
        if (area.schedules.every(schedule => { return schedule.dateStrings.length != 0 })) return;
    
        // 収集日のCVSを取得（await）
        const csvTexts = await Promise.all(area.schedules.map(schedule => {
            return app.getCSV('data/schedules/' + schedule.file);
        }));

        // CSVを日付文字列（Data.toLocalDateString()）に変換して変数に保存
        csvTexts.forEach((csvText, i) => {
            
            // csvを日付配列に変換
            // areaオブジェクトの中に保存
            area.schedules[i].dateStrings = csvText.split('\r\n').map(text => {
                return new Date(text).toLocaleDateString();
            });
        });
    };

    /**
     * 地区select要素の初期化。
     * 
     * @param {object[]} areas areaオブジェクトの配列。
     */
    const initCalendarArea = () => {

        const areas = app.areas;
        const select = document.getElementById('calendar-area-select');

        // 地区の選択項目を生成
        html = '';
        areas.forEach(area => {
            html += `<option value="${area.id}" class="calendar-area-item">${area.name}</option>`;
        });
        select.innerHTML = html;

        // 地区選択時のイベントをセット。
        select.addEventListener('change', async (e) => {

            // 選択中のoption要素にクラスを付与
            document.querySelectorAll('.calendar-area-item--selected').forEach(item => {
                item.classList.remove('calendar-area-item--selected');
            });
            document.querySelector(`.calendar-area-item[value="${e.currentTarget.value}"]`).classList.add('calendar-area-item--selected');

            // 収集日データの読み込みとカレンダー更新
            const area = areas.find(item => item.id == e.currentTarget.value);
            await loadSchedules(area);
            updateCalendar();
        });
    
    };

    /**
     * カレンダーの月コントロール要素の初期化。
     * 
     * @param {number} year カレンダーの年を西暦で表す4桁の数値、指定しない場合は今日の年。
     * @param {number} month カレンダーの月を表すの数値（1 - 12）、指定しない場合は今日の月。
     */
    const initCalnedarControl = (year = new Date().getFullYear(), month = new Date().getMonth() + 1) => {

        const control = document.querySelector('.calendar-control');

        // 年月テキストの更新メソッド
        const updateYearMonth = (year, month) => {
            control.dataset.year = year;
            control.dataset.month = month;
            document.querySelector('.calendar-control-year').innerText = year + '年';
            document.querySelector('.calendar-control-month').innerText = month + '月';
        };
        updateYearMonth(year, month);

        // prevボタンのイベントをセット
        document.querySelector('.calendar-control-prev').addEventListener('click', e => {
            if (control.dataset.month == 1) {
                updateYearMonth(--control.dataset.year, 12);
            } else {
                updateYearMonth(control.dataset.year, --control.dataset.month);
            }
            updateCalendar();
        });

        // nextボタンのイベントをセット
        document.querySelector('.calendar-control-next').addEventListener('click', e => {
            if (control.dataset.month == 12) {
                updateYearMonth(++control.dataset.year, 1);
            } else {
                updateYearMonth(control.dataset.year, ++control.dataset.month);
            }
            updateCalendar();
        });

    };

    /**
     * カレンダーテーブルを更新する。
     * 
     */
    const updateCalendar = () => {

        const area = app.areas.find(area => {
            return area.id == document.querySelector('.calendar-area-select').value;
        });

        const year = document.querySelector('.calendar-control').dataset.year;
        const month = document.querySelector('.calendar-control').dataset.month;
        let date = new Date(`${year}/${month}/1`);

        const today = new Date();

        const fragment = document.createDocumentFragment();

        // カレンダーの日付作成処理
        // fragment 使う
        date.setDate(1 - date.getDay());
        while (fragment.childElementCount < 2 || (month - 1) == date.getMonth()) {
            let tr = document.createElement('tr');
            tr.classList.add('calendar-table-row');
            for (let i = 0; i < 7; i++) {
                let td = document.createElement('td');
                td.innerHTML = `<span class="calendar-table-date-num">${date.getDate()}</span>`;
                td.dataset.Date = date.toLocaleDateString();
                td.classList.add('calendar-table-date');
                
                // 曜日等のクラス付与
                if (td.dataset.Date == today.toLocaleDateString()) td.classList.add('calendar-table-date--today');
                if (date.getDay() == 0) td.classList.add('calendar-table-date--sun');
                if (date.getDay() == 6) td.classList.add('calendar-table-date--sat');
                if (date.getMonth() != month - 1) td.classList.add('calendar-table-date--alt');
                
                // 収集日のテキストとクラス付与
                area.schedules.forEach(schedule => {

                    if (schedule.dateStrings.includes(date.toLocaleDateString())) {
                        const sortKey = app.sortKey(schedule.sort);
                        const span = document.createElement('span');
                        span.classList.add('calendar-table-date-sort', `calendar-table-date-sort--${sortKey}`);
                        span.innerText = schedule.sort.replace('ごみ', '');
                        td.appendChild(span);
                        td.classList.add(`calendar-table-date--${sortKey}`);
                    }
                });

                tr.appendChild(td);
                date.setDate(date.getDate() + 1);
            };
            fragment.appendChild(tr);

            /* 万が一の無限ループ対策 */
            if (fragment.childElementCount > 10) break;
        };
        // DOMに追加
        document.getElementById('calendar-table-body').replaceChildren(fragment);
    };

    /**
     * ##############
     *  処理実行
     * ##############
     */

    // 地区のcsvデータ取得
    app.areas = await parseCalendarCsv('data/area.csv');

    // 地区選択(select要素)の初期化
    initCalendarArea();

    // カレンダーコントロールの初期化
    initCalnedarControl();

    // localStorage による地区の自動選択
    const select = document.getElementById('calendar-area-select');
    select.addEventListener('change', e => {
        window.localStorage.areaid = select.value;
    });
    if (window.localStorage.hasOwnProperty('areaid')) {
        select.value = window.localStorage.areaid;
    };
    select.dispatchEvent(new Event('change'));

};

/**
 * 収集拠点ページの初期化メソッド
 * leaflet.js が必要
 */
app.initPlace = async () => {

    /**
     * ##############
     *  関数定義
     * ##############
     */

    app.places;
    app.placeCategories;
    app.placeMap;
    app.placeLayers;
    app.placeCurrentLayer;

    const getIconOptions = (index) => {
        return [
            // index: 0
            {
                iconUrl: 'marker-icon.png',
                iconRetinaUrl: 'marker-icon-2x.png'
            },
            // index: 1
            {
                iconUrl: 'marker-icon-green.png',
                iconRetinaUrl: 'marker-icon-green-2x.png'
            },
            // index: 2
            {
                iconUrl: 'marker-icon-red.png',
                iconRetinaUrl: 'marker-icon-red-2x.png'
            },
            // index: 3
            {
                iconUrl: 'marker-icon-purple.png',
                iconRetinaUrl: 'marker-icon-purple-2x.png'
            },
            // index: 4
            {
                iconUrl: 'marker-icon-yellow.png',
                iconRetinaUrl: 'marker-icon-yellow-2x.png'
            },
            // index: 5
            {
                iconUrl: 'marker-icon-orange.png',
                iconRetinaUrl: 'marker-icon-orange-2x.png'
            }
        ][index];
    };

    /**
     * place.csv を読み込んでオブジェクトに保存。
     * 
     * @param {string} path place.csv の場所。
     * @return {object[]} placeオブジェクトの配列。
     */
    const parsePlaceCsv = async path => {
        const csv = await app.getCSV(path);
        const places = [];
        app.readCSV(csv, 1, (row, values) => {
            places.push({
                category    : values[0],
                name        : values[1],
                address     : values[2],
                tel         : values[3],
                description : values[4],
                linkTitle   : values[5],
                linkUrl     : values[6],
                latlng      : { lat: values[7], lng: values[8] }
            });
        });
        return places;
    };

    /**
     * 種類selectの初期化。
     * 
     * @param {object[]} places 
     */
    const initCategory = () => {

        const select = document.querySelector('.place-category-select');

        // 項目の追加
        const htmls = app.placeCategories.map(category => {
            return `<option value="${category}" class="place-category-item">${category}</option>`;
        });
        select.innerHTML = htmls.join('');

        // 変更時イベント
        select.addEventListener('change', e => {
            
            // 選択中のoption要素にクラスを付与
            document.querySelectorAll('.place-category-item--selected').forEach(item => {
                item.classList.remove('place-category-item--selected');
            });
            document.querySelector(`.place-category-item[value="${e.currentTarget.value}"]`).classList.add('place-category-item--selected');

            // LayerGroup 切り替え
            if (app.placeCurrentLayer) app.placeMap.removeLayer(app.placeCurrentLayer);
            app.placeCurrentLayer = app.placeLayers[e.currentTarget.value];
            app.placeMap.addLayer(app.placeCurrentLayer);

        });
    };

    /**
     * 種類ごとの marker を持った LayerGroup の作成
     * @return {LayerGroup[]}
     */
    const createMarkerLayers = () => {

        // marker を入れる LayerGroup を生成
        const layers = {};
        app.placeCategories.forEach(category => {
            layers[category] =  L.layerGroup();
        });

        // marker を作成して LayerGroup に追加
        app.places.forEach(place => {
            if (!layers.hasOwnProperty(place.category)) return;

            // category 毎にマーカーアイコンを変更
            const icon = new L.Icon.Default();
            Object.assign(icon.options, getIconOptions(app.placeCategories.indexOf(place.category) % 6));

            // マーカー生成
            const marker = L.marker(place.latlng, {
                title: place.name,
                alt: place.name,
                riseOnHover: true,
                icon: icon
            });
            marker.bindTooltip(place.name);
            marker.bindPopup(`
                <h2 class="place-popup-name">${place.name}</h2>
                <p class="place-popup-address"><span class="place-popup-address-attr">住所:</span>${place.address}</p>
                <p class="place-popup-tel"><span class="place-popup-tel-attr">電話:</span><a href="tel:${place.tel}">${place.tel}</a></p>
                <p class="place-popup-description">${place.description}</p>
                <p class="place-popup-link"><a href="${place.linkUrl}" target="blank">${place.linkTitle}</a></p>
            `);
            marker.addTo(layers[place.category]);
            
            // place オブジェクトに marker への参照を追加
            place.marker = marker;
        });
        return layers;
    };

    // L.Map インスタンスを生成
    const createMap = (elementId) => {
        
        // TileLayer の生成
        const tileLayers = {
            // 標準
            '標準' : L.tileLayer('https://cyberjapandata.gsi.go.jp/xyz/std/{z}/{x}/{y}.png', {
                attribution: "<a href='https://maps.gsi.go.jp/development/ichiran.html' target='_blank'>地理院タイル</a>"
            }),
            // 淡色
            '淡色' : L.tileLayer('https://cyberjapandata.gsi.go.jp/xyz/pale/{z}/{x}/{y}.png', {
                attribution: "<a href='https://maps.gsi.go.jp/development/ichiran.html' target='_blank'>地理院タイル</a>"
            })
        };

        const map = L.map(elementId, {
            center: [34.577256354814104, 135.6296539306641],
            zoom: 13,
            layers: tileLayers['淡色']
        });

        // Zoom Control の位置を調整 
        map.zoomControl.setPosition('bottomleft');

        // Layer Control 追加
        const layerControl = L.control.layers(tileLayers).addTo(map);
        layerControl.setPosition('bottomleft');

        return map;
    };

    /**
     * ##############
     *  処理実行
     * ##############
     */

    // place.csv のデータを読み込み
    app.places = await parsePlaceCsv('./data/place.csv');
    app.placeCategories = [...new Set(app.places.map(p => { return p.category}))];

    // Leaflet ライブラリの読み込み
    if (typeof L == 'undefined') {
        app.import('lib/leaflet/leaflet.css');
        await app.import('lib/leaflet/leaflet.js');
    }

    // 種類selectの初期化
    initCategory();

    // map オブジェクト生成
    app.placeLayers = createMarkerLayers();
    app.placeMap = createMap('place-map');

    // 種類 select のクリックイベントを実行
    document.querySelector('.place-category-select').dispatchEvent(new Event('change'));

    // 一覧ボタンのイベントをセット
    document.querySelector('.place-category-list').addEventListener('click', e => {
        
        // 選択中の種類と一致する place オブジェクトを取得
        let htmls = [];
        const altPlaces = app.places.forEach((place, i) => {
            const category = document.querySelector('.place-category-select').value;
            if (place.category != category) return;
            htmls.push(`<div class="dialog-place-item" data-index="${i}">${place.name}</div>`);
        });
        app.showDialog(htmls.join(''));

        // 一覧の項目クリック時にイベントをセット
        document.querySelectorAll('.dialog-place-item').forEach(item => {
            item.addEventListener('click', e => {
                const marker = app.places[e.currentTarget.dataset.index].marker;
                // 一覧ダイアログを閉じる
                app.closeDialog();
                // マーカーを中心に表示して Tooltip も表示
                app.placeMap.panTo(marker.getLatLng());
                marker.toggleTooltip();
            });
        });
    });
};

/**
 * 空中写真ページの初期化メソッド
 * leaflet.js が必要
 */
app.initGsi = async () => {

    app.gsiCurrentLayer;

    // leaflet ライブラリ読み込み
    if (typeof L == 'undefined') {
        app.import('lib/leaflet/leaflet.css');
        await app.import('lib/leaflet/leaflet.js');
    }    

    // 標準地図
    const tileLayer = L.tileLayer('https://cyberjapandata.gsi.go.jp/xyz/std/{z}/{x}/{y}.png', {
        attribution: "<a href='https://maps.gsi.go.jp/development/ichiran.html' target='_blank'>地理院タイル</a>"
    });

    // 航空写真
    const photoLayers = {
        // 2020?
        '現在' : L.tileLayer('https://cyberjapandata.gsi.go.jp/xyz/seamlessphoto/{z}/{x}/{y}.jpg', {
            attribution: "<a href='https://maps.gsi.go.jp/development/ichiran.html' target='_blank'>地理院タイル</a>",
            minZoom: 10
        }),

        // 1984-1986
        '1985年頃' : L.tileLayer('https://cyberjapandata.gsi.go.jp/xyz/gazo3/{z}/{x}/{y}.jpg', {
            attribution: "<a href='https://maps.gsi.go.jp/development/ichiran.html' target='_blank'>地理院タイル</a>",
            maxZoom: 17,
            minZoom: 10
        }),

        // 1974-1978
        '1975年頃' : L.tileLayer('https://cyberjapandata.gsi.go.jp/xyz/gazo1/{z}/{x}/{y}.jpg', {
            attribution: "<a href='https://maps.gsi.go.jp/development/ichiran.html' target='_blank'>地理院タイル</a>",
            maxZoom: 17,
            minZoom: 10
        }),

        // 1961-1969
        '1965年頃' : L.tileLayer('https://cyberjapandata.gsi.go.jp/xyz/ort_old10/{z}/{x}/{y}.png', {
            attribution: "<a href='https://maps.gsi.go.jp/development/ichiran.html' target='_blank'>地理院タイル</a>",
            maxZoom: 17,
            minZoom: 10
        }),

        // 1945-1950
        '1950年頃' : L.tileLayer('https://cyberjapandata.gsi.go.jp/xyz/ort_USA10/{z}/{x}/{y}.png', {
            attribution: "<a href='https://maps.gsi.go.jp/development/ichiran.html' target='_blank'>地理院タイル</a>",
            maxZoom: 17,
            minZoom: 10
        }),

        // 1936-1942
        '1940年頃' : L.tileLayer('https://cyberjapandata.gsi.go.jp/xyz/ort_riku10/{z}/{x}/{y}.png', {
            attribution: "<a href='https://maps.gsi.go.jp/development/ichiran.html' target='_blank'>地理院タイル</a>",
            maxZoom: 18,
            minZoom: 13
        }),

        // 地図のみ
        '標準地図' : undefined
    };

    let currentLayer = photoLayers['現在'];
    const map = app.gsiMap = L.map('gsi-map', {
        center: [34.577256354814104, 135.6296539306641],
        zoom: 14,
        layers: [tileLayer, currentLayer]
    });

    const select = document.getElementById('gsi-overlay-select');

    // Zoom Control の位置を調整 
    map.zoomControl.setPosition('bottomleft');

    // 年代 select の項目を追加
    let htmls = [];
    Object.keys(photoLayers).forEach(key => {
        htmls.push(`<option value=${key}>${key}</option>`);
    });
    select.innerHTML = htmls.join('');

    // 年代 select のイベントをセット
    select.addEventListener('change', e => {
        const selectLayer = photoLayers[e.currentTarget.value];
        if (selectLayer) map.addLayer(selectLayer);
        if (currentLayer) map.removeLayer(currentLayer);
        currentLayer = selectLayer;
    });

};  


/**
 * ####################
 *  メイン処理の実行部分
 * ####################
 */
window.addEventListener('DOMContentLoaded', (e) => {

    // ダイアログを初期化
    app.initDialog();

    // メニューを初期化
    app.initMenu();

    // 初期表示コンテンツの初期化（url.hash"#"で指定）
    const menuItem = document.querySelector(`.header-menu-item[data-content-id="${location.hash.replace('#', '')}"]`) || document.querySelector('.header-menu-item');
    menuItem.dispatchEvent(new Event('click'));

});